# From test_grammar.py
# RUNNABLE!
def check_syntax_error(statement):
    try:
        compile(statement, '<test string>', 'exec')
    except SyntaxError:
        return
    assert False

def test_yield():
    # Requires parentheses as call argument
    def g():
        f((yield 1), 1)
    def g():
        f((yield from ()))
    def g():
        f((yield from ()), 1)
    def g():
        f((yield 1))

    # Allowed as standalone statement
    def g():
        yield 1

    def g():
        yield from ()

    # Allowed as RHS of assignment
    def g():
        x = yield 1

    def g():
        x = yield from ()

    # Ordinary yield accepts implicit tuples
    def g():
        yield 1, 1

    def g():
        x = yield 1, 1

    # 'yield from' does not
    check_syntax_error("def g(): yield from (), 1")
    check_syntax_error("def g(): x = yield from (), 1")
    # Requires parentheses as subexpression
    def g():
        1, (yield 1)

    def g():
        1, (yield from ())

    check_syntax_error("def g(): 1, yield 1")
    check_syntax_error("def g(): 1, yield from ()")
    # Requires parentheses as call argument
    def g():
        f((yield 1))

    def g():
        f((yield 1), 1)

    def g():
        f((yield from ()))

    def g():
        f((yield from ()), 1)

    check_syntax_error("def g(): f(yield 1)")
    check_syntax_error("def g(): f(yield 1, 1)")
    check_syntax_error("def g(): f(yield from ())")
    check_syntax_error("def g(): f(yield from (), 1)")
    # Not allowed at top level
    check_syntax_error("yield")
    check_syntax_error("yield from")
    # Not allowed at class scope
    check_syntax_error("class foo:yield 1")
    check_syntax_error("class foo:yield from ()")
    # Check annotation refleak on SyntaxError
    check_syntax_error("def g(a:(yield)): pass")

test_yield()

# From test_types.py
# Bug was needing parens around (yield 2)
def gen_func():
    yield 1
    return (yield 2)
gen = gen_func()



"""
CPython 3.12 bytecode opcodes

This is like Python 3.12's opcode.py  with some classification
of stack usage and information for formatting instructions.
"""

import xdis.opcodes.opcode_311 as opcode_311
from xdis.opcodes.base import (
    binary_op,
    def_op,
    finalize_opcodes,
    init_opdata,
    jrel_op,
    name_op,
    rm_op,
    update_pj3,
)
from xdis.opcodes.opcode_311 import opcode_arg_fmt311, opcode_extended_fmt311

version_tuple = (3, 12)
python_implementation = "CPython"

loc = locals()

init_opdata(loc, opcode_311, version_tuple)


# extend opcodes to cover pseudo ops
loc["opname"].extend([f"<{i}>" for i in range(256, 267)])
loc["oppop"].extend([0] * 11)
loc["oppush"].extend([0] * 11)

# fmt: off
## These are removed / replaced since 3.11...
#           OP NAME                      OPCODE
#----------------------------------------------
rm_op(loc, "UNARY_POSITIVE"               , 10)
rm_op(loc, "PRINT_EXPR"                   , 70)

rm_op(loc, "LIST_TO_TUPLE"                , 82)
rm_op(loc, "IMPORT_STAR"                  , 84)
rm_op(loc, "ASYNC_GEN_WRAP"               , 87)
rm_op(loc, "PREP_RERAISE_STAR"            , 88)

rm_op(loc, "JUMP_IF_FALSE_OR_POP"         , 111)
rm_op(loc, "JUMP_IF_TRUE_OR_POP"          , 112)
rm_op(loc, "POP_JUMP_FORWARD_IF_FALSE"    , 114)
rm_op(loc, "POP_JUMP_FORWARD_IF_TRUE"     , 115)

rm_op(loc, "POP_JUMP_FORWARD_IF_NONE"     , 129)
rm_op(loc, "POP_JUMP_FORWARD_IF_NOT_NONE" , 128)
rm_op(loc, "LOAD_CLASSDEREF"              , 148)

rm_op(loc, "PRECALL"                      , 166)
rm_op(loc, "POP_JUMP_BACKWARD_IF_NOT_NONE", 173)
rm_op(loc, "POP_JUMP_BACKWARD_IF_NONE"    , 174)
rm_op(loc, "POP_JUMP_BACKWARD_IF_FALSE"   , 175)
rm_op(loc, "POP_JUMP_BACKWARD_IF_TRUE"    , 176)

rm_op(loc, "YIELD_VALUE"                  , 86)
rm_op(loc, "LOAD_METHOD"                  , 160)

## These are new since 3.11...
#            OP NAME                              OPCODE POP PUSH
#---------------------------------------------------------------
def_op(loc    , "INTERPRETER_EXIT"                 , 3  ,   1, 0)
def_op(loc    , "END_FOR"                          , 4  ,   2, 0)
def_op(loc    , "END_SEND"                         , 5  ,   1, 0)
def_op(loc    , "RESERVED"                         , 17 ,   0, 0)

binary_op(loc , "BINARY_SLICE"                     , 26 ,   2, 0)
binary_op(loc , "STORE_SLICE"                      , 27 ,   4, 0)

def_op(loc    , "CLEANUP_THROW"                    , 55 ,   2, 1)
def_op(loc    , "LOAD_LOCALS"                      , 87 ,   0, 1)
def_op(loc    , "RETURN_CONST"                     , 121,   0, 0)
def_op(loc    , "LOAD_FAST_CHECK"                  , 127,   0, 1)

jrel_op(loc   , "POP_JUMP_IF_FALSE"                , 114,   1, 0)
jrel_op(loc   , "POP_JUMP_IF_TRUE"                 , 115,   1, 0)
jrel_op(loc   , "POP_JUMP_IF_NOT_NONE"             , 128,   1, 0)
jrel_op(loc   , "POP_JUMP_IF_NONE"                 , 129,   1, 0)

def_op(loc    , "LOAD_SUPER_ATTR"                  , 141,   3, 1)
def_op(loc    , "LOAD_FAST_AND_CLEAR"              , 143,   0, 1)
def_op(loc    , "YIELD_VALUE"                      , 150,   1, 1)
def_op(loc    , "CALL_INTRINSIC_1"                 , 173,   1, 1)
def_op(loc    , "CALL_INTRINSIC_2"                 , 174,   2, 1)
def_op(loc    , "LOAD_FROM_DICT_OR_GLOBALS"        , 175,   1, 1)
def_op(loc    , "LOAD_FROM_DICT_OR_DEREF"          , 176,   1, 1)

#            OP NAME                               OPCODE POP PUSH
def_op(loc    , "INSTRUMENTED_LOAD_SUPER_ATTR"     , 237,   2, 0)
def_op(loc    , "INSTRUMENTED_POP_JUMP_IF_NONE"    , 238,   1, 1)
def_op(loc    , "INSTRUMENTED_POP_JUMP_IF_NOT_NONE", 239,   1, 1)
def_op(loc    , "INSTRUMENTED_RESUME"              , 240,   1, 1)
def_op(loc    , "INSTRUMENTED_CALL"                , 241,   1, 1)
def_op(loc    , "INSTRUMENTED_RETURN_VALUE"        , 242,   1, 0)
def_op(loc    , "INSTRUMENTED_YIELD_VALUE"         , 243,   1, 1)
def_op(loc    , "INSTRUMENTED_CALL_FUNCTION_EX"    , 244,   1, 1)
def_op(loc    , "INSTRUMENTED_JUMP_FORWARD"        , 245,   1, 1)
def_op(loc    , "INSTRUMENTED_JUMP_BACKWARD"       , 246,   1, 1)
def_op(loc    , "INSTRUMENTED_RETURN_CONST"        , 247,   1, 1)
def_op(loc    , "INSTRUMENTED_FOR_ITER"            , 248,   1, 1)
def_op(loc    , "INSTRUMENTED_POP_JUMP_IF_FALSE"   , 249,   1, 1)
def_op(loc    , "INSTRUMENTED_POP_JUMP_IF_TRUE"    , 250,   1, 1)
def_op(loc    , "INSTRUMENTED_END_FOR"             , 251,   2, 0)
def_op(loc    , "INSTRUMENTED_END_SEND"            , 252,   1, 0)
def_op(loc    , "INSTRUMENTED_INSTRUCTION"         , 253,   1, 1)
def_op(loc    , "INSTRUMENTED_LINE"                , 254,   1, 1)

#            OP NAME                              OPCODE POP PUSH
def_op(loc    , "SETUP_FINALLY"                    , 256,   0, 1)
def_op(loc    , "SETUP_CLEANUP"                    , 257,   0, 1)
def_op(loc    , "SETUP_WITH"                       , 258,   0, 1)
def_op(loc    , "POP_BLOCK"                        , 259,   0, 1)

jrel_op(loc   , "JUMP"                             , 260,   0, 0)
jrel_op(loc   , "JUMP_NO_INTERRUPT"                , 261,   0, 0)

name_op(loc   , "LOAD_METHOD"                      , 262,   0, 1)
def_op(loc    , "LOAD_SUPER_METHOD"                , 263,   1, 0)
def_op(loc    , "LOAD_ZERO_SUPER_METHOD"           , 264,   1, 0)
def_op(loc    , "LOAD_ZERO_SUPER_ATTR"             , 265,   1, 0)
def_op(loc    , "STORE_FAST_MAYBE_NULL"            , 266,   1, 0)

### update opinfo tables
loc["hasnargs"] = []
loc["hasstore"] = []
loc["hasvargs"] = []

loc["hasconst"].append(121)
loc["hasfree"].extend([148, 176])
# hasjrel removed 111 112 173 174 175 176 and added 260 261
loc["hasjrel"] = [93, 110, 114, 115, 123, 128, 129, 134, 140, 260, 261]
loc["haslocal"].extend([127, 143, 266])
# hasname removed 160 and added 141 175 262 234 264 265
loc["hasname"] = [90, 91, 95, 96, 97, 98, 101, 106, 108, 109, 116, 141, 175, 262, 263, 264, 265]
# new hasarg table
loc.update({"hasarg": [90, 91, 92, 93, 94, 95, 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110,\
                       114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127, 128, 129, 130, 131, 132,
                       133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 149, 150, 151, 152,
                       155, 156, 157, 162, 163, 164, 165, 171, 172, 173, 174, 175, 176, 237, 238, 239, 240, 241, 242,
                       243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 260, 261, 262, 263, 264, 265, 266]})

# new hasexc table
loc.update({"hasexc":[264, 265, 266]})

### add new arg formats
_intrinsic_1_descs = [
    "INTRINSIC_1_INVALID",
    "INTRINSIC_PRINT",
    "INTRINSIC_IMPORT_STAR",
    "INTRINSIC_STOPITERATION_ERROR",
    "INTRINSIC_ASYNC_GEN_WRAP",
    "INTRINSIC_UNARY_POSITIVE",
    "INTRINSIC_LIST_TO_TUPLE",
    "INTRINSIC_TYPEVAR",
    "INTRINSIC_PARAMSPEC",
    "INTRINSIC_TYPEVARTUPLE",
    "INTRINSIC_SUBSCRIPT_GENERIC",
    "INTRINSIC_TYPEALIAS",
]

_intrinsic_2_descs = [
    "INTRINSIC_2_INVALID",
    "INTRINSIC_PREP_RERAISE_STAR",
    "INTRINSIC_TYPEVAR_WITH_BOUND",
    "INTRINSIC_TYPEVAR_WITH_CONSTRAINTS",
    "INTRINSIC_SET_FUNCTION_TYPE_PARAMS",
]

def format_CALL_INTRINSIC_1(arg) -> str:
    return _intrinsic_1_descs[arg]

def format_CALL_INTRINSIC_2(arg) -> str:
    return _intrinsic_2_descs[arg]


opcode_extended_fmt = opcode_extended_fmt312 = opcode_extended_fmt311.copy()
opcode_arg_fmt = opcode_arg_fmt12 = opcode_arg_fmt311.copy()

### update arg formatting
opcode_arg_fmt312 = {
    **opcode_arg_fmt311,
    **{
        "CALL_INTRINSIC_1": format_CALL_INTRINSIC_1,
        "CALL_INTRINSIC_2": format_CALL_INTRINSIC_2,
    },
}

from xdis.opcodes.opcode_311 import findlinestarts

update_pj3(globals(), loc)
finalize_opcodes(loc)



# (C) Copyright 2018, 2020-2024 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""CPython core set of bytecode opcodes based on version 2.3

This is used in bytecode disassembly among other things. This is
similar to the opcodes in Python's opcode.py library.

If this file changes the other opcode files may have to be adjusted accordingly.
"""

from typing import Optional, Tuple

from xdis.opcodes.base import (
    binary_op,
    call_op,
    compare_op,
    const_op,
    def_op,
    free_op,
    init_opdata,
    jabs_op,
    jrel_op,
    local_op,
    name_op,
    nargs_op,
    store_op,
    ternary_op,
    unary_op,
    varargs_op,
)
from xdis.opcodes.format.basic import format_MAKE_FUNCTION_10_27, opcode_arg_fmt_base
from xdis.opcodes.format.extended import (
    get_arglist,
    get_instruction_arg,
    opcode_extended_fmt_base,
)

loc = locals()
init_opdata(loc, None, None)

# Instruction opcodes for compiled code
# Blank lines correspond to available opcodes

# If the POP field is -1 and the opcode is var args operation
# (hasvargs | hasnargs) operation, then
# the operand holds the size.

# fmt: off
#          OP NAME              OPCODE POP PUSH
#-----------------------------------------------
def_op(loc, "STOP_CODE",               0,  0,  0, fallthrough=False)
def_op(loc, "POP_TOP",                 1,  1,  0)
def_op(loc, "ROT_TWO",                 2,  2,  2)
def_op(loc, "ROT_THREE",               3,  3,  3)
def_op(loc, "DUP_TOP",                 4,  0,  1)
def_op(loc, "ROT_FOUR",                5,  4,  4)

unary_op(loc, "UNARY_POSITIVE",          10)
unary_op(loc, "UNARY_NEGATIVE",          11)
unary_op(loc, "UNARY_NOT",               12)
unary_op(loc, "UNARY_CONVERT",           13)

unary_op(loc, "UNARY_INVERT",            15)

binary_op(loc, "BINARY_POWER",           19)

binary_op(loc, "BINARY_MULTIPLY",        20)
binary_op(loc, "BINARY_DIVIDE",          21)
binary_op(loc, "BINARY_MODULO",          22)
binary_op(loc, "BINARY_ADD",             23)
binary_op(loc, "BINARY_SUBTRACT",        24)
binary_op(loc, "BINARY_SUBSCR",          25)
binary_op(loc, "BINARY_FLOOR_DIVIDE",    26)
binary_op(loc, "BINARY_TRUE_DIVIDE",     27)
binary_op(loc, "INPLACE_FLOOR_DIVIDE",   28)
binary_op(loc, "INPLACE_TRUE_DIVIDE",    29)

unary_op(loc, "SLICE+0",                 30)
binary_op(loc, "SLICE+1",                31)
binary_op(loc, "SLICE+2",                32)
ternary_op(loc, "SLICE+3",               33)

#          OP NAME              OPCODE POP PUSH
#-----------------------------------------------
store_op(loc, "STORE_SLICE+0",        40,  2,  0)
store_op(loc, "STORE_SLICE+1",        41,  3,  0)
store_op(loc, "STORE_SLICE+2",        42,  3,  0)
store_op(loc, "STORE_SLICE+3",        43,  4,  0)

def_op(loc, "DELETE_SLICE+0",         50,  1,  0)
def_op(loc, "DELETE_SLICE+1",         51,  2,  0)
def_op(loc, "DELETE_SLICE+2",         52,  2,  0)
def_op(loc, "DELETE_SLICE+3",         53,  3,  0)

binary_op(loc, "INPLACE_ADD",         55)
binary_op(loc, "INPLACE_SUBTRACT",    56)
binary_op(loc, "INPLACE_MULTIPLY",    57)
binary_op(loc, "INPLACE_DIVIDE",      58)
binary_op(loc, "INPLACE_MODULO",      59)
store_op(loc, "STORE_SUBSCR",         60,  3,  0) # Implements TOS1[TOS] = TOS2.
def_op(loc, "DELETE_SUBSCR",          61,  2,  0) # Implements del TOS1[TOS].

binary_op(loc, "BINARY_LSHIFT",       62)
binary_op(loc, "BINARY_RSHIFT",       63)
binary_op(loc, "BINARY_AND",          64)
binary_op(loc, "BINARY_XOR",          65)
binary_op(loc, "BINARY_OR",           66)
binary_op(loc, "INPLACE_POWER",       67)

def_op(loc, "GET_ITER",               68,  1,  1)

def_op(loc, "PRINT_EXPR",             70,  1,  0)
def_op(loc, "PRINT_ITEM",             71,  1,  0)
def_op(loc, "PRINT_NEWLINE",          72,  0,  0)
def_op(loc, "PRINT_ITEM_TO",          73,  2,  0)
def_op(loc, "PRINT_NEWLINE_TO",       74,  1,  0)
def_op(loc, "INPLACE_LSHIFT",         75,  2,  1)
def_op(loc, "INPLACE_RSHIFT",         76,  2,  1)
def_op(loc, "INPLACE_AND",            77,  2,  1)
def_op(loc, "INPLACE_XOR",            78,  2,  1)
def_op(loc, "INPLACE_OR",             79,  2,  1)
def_op(loc, "BREAK_LOOP",             80,  0,  0, fallthrough=False)

def_op(loc, "LOAD_LOCALS",            82,  0,  1)
def_op(loc, "RETURN_VALUE",           83,  1,  0, fallthrough=False)
def_op(loc, "IMPORT_STAR",            84,  1,  0)
def_op(loc, "EXEC_STMT",              85,  3,  0)
def_op(loc, "YIELD_VALUE",            86,  1,  1)

def_op(loc, "POP_BLOCK",              87,  0,  0)
def_op(loc, "END_FINALLY",            88,  1,  0)
def_op(loc, "BUILD_CLASS",            89,  2,  0)

HAVE_ARGUMENT = 90              # Opcodes from here have an argument:

#             OP NAME              OPCODE POP PUSH
#-----------------------------------------------
store_op(loc, "STORE_NAME",            90,  1,  0, is_type="name")
                                                    # Operand is in name list
name_op(loc, "DELETE_NAME",            91,  0,  0)  # ""
varargs_op(loc, "UNPACK_SEQUENCE",     92, -1,  1)  # TOS is number of tuple items
jrel_op(loc, "FOR_ITER",               93,  0,  1)  # TOS is read

store_op(loc, "STORE_ATTR",            95,  2,  0, is_type="name")
                                                    # Operand is in name list
name_op(loc, "DELETE_ATTR",            96,  1,  0)  # ""
store_op(loc, "STORE_GLOBAL",          97,  1,  0, is_type="name")  # ""
name_op(loc, "DELETE_GLOBAL",          98,  0,  0)  # ""
nargs_op(loc, "DUP_TOPX",              99, -1,  2)  # number of items to duplicate

const_op(loc, "LOAD_CONST",           100,  0,  1)  # Operand is in const list
loc["nullaryloadop"].add(100)

loc["nullaryloadop"].add(100)

name_op(loc, "LOAD_NAME",             101,  0,  1)  # Operand is in name list
loc["nullaryloadop"].add(101)

varargs_op(loc, "BUILD_TUPLE",        102, -1,  1)  # TOS is number of tuple items
varargs_op(loc, "BUILD_LIST",         103, -1,  1)  # TOS is number of list items
varargs_op(loc, "BUILD_MAP",          104,  0,  1)  # TOS is number of kwarg items.
                                                  # Always zero for now
name_op(loc, "LOAD_ATTR",             105,  1,  1)  # Operand is in name list
compare_op(loc, "COMPARE_OP",         106,  2,  1)  # Comparison operator

name_op(loc, "IMPORT_NAME",           107,  0,  1)  # For < 2.6;  Imports namei; module
                                                    # pushed
name_op(loc, "IMPORT_FROM",           108,  0,  1)  # Operand is in name list

jrel_op(loc, "JUMP_FORWARD",          110,  0,  0, fallthrough=False)
                                                # Number of bytes to skip
jrel_op(loc, "JUMP_IF_FALSE",         111,  1,  1, True)  # ""

jrel_op(loc, "JUMP_IF_TRUE",          112,  1,  1, True)  # ""
jabs_op(loc, "JUMP_ABSOLUTE",         113,  0,  0, fallthrough=False)
                                                # Target byte offset from beginning of
                                                 #code

name_op(loc, "LOAD_GLOBAL",           116,  0,  1)  # Operand is in name list
loc["nullaryloadop"].add(116)


jabs_op(loc, "CONTINUE_LOOP",         119,  0,  0, fallthrough=False)  # Target address
jrel_op(loc, "SETUP_LOOP",            120,  0,  0, conditional=True)  # Distance to
                                                                      # target address
jrel_op(loc, "SETUP_EXCEPT",          121,  0,  3, conditional=True)  # ""
jrel_op(loc, "SETUP_FINALLY",         122,  0,  3, conditional=True)  # ""

local_op(loc, "LOAD_FAST",            124,  0,  1)  # Local variable number
loc["nullaryloadop"].add(124)

store_op(loc, "STORE_FAST",           125,  1,  0, is_type="local")  # Local variable
                                                                     # number
local_op(loc, "DELETE_FAST",          126,  0,  0) # Local variable number is in operand

nargs_op(loc, "RAISE_VARARGS",        130, -1,  2, fallthrough=False)
                                                # Number of raise arguments (1, 2, or 3)
call_op(loc, "CALL_FUNCTION",        131, -1,  2)  # TOS is #args + (#kwargs << 8)

nargs_op(loc, "MAKE_FUNCTION",        132, -1,  2)  # TOS is number of args with
                                                    # default values
varargs_op(loc, "BUILD_SLICE",        133,  2,  1)  # TOS is number of items

def_op(loc, "MAKE_CLOSURE",           134, -3,  1)

free_op(loc, "LOAD_CLOSURE",          135,  0,  1)  # Load of a closured variable
loc["nullaryloadop"].add(135)

free_op(loc, "LOAD_DEREF",            136,  0,  1)
loc["nullaryop"].add(136)
loc["nullaryloadop"].add(136)

store_op(loc, "STORE_DEREF",          137,  1,  0, is_type="free")

call_op(loc, "CALL_FUNCTION_VAR",    140, -2,  1) # #args + (#kwargs << 8)
call_op(loc, "CALL_FUNCTION_KW",     141, -2,  1) # #args + (#kwargs << 8)
call_op(loc, "CALL_FUNCTION_VAR_KW", 142, -3,  1) # #args + (#kwargs << 8)

def_op(loc, "EXTENDED_ARG", 143)
# fmt: on

EXTENDED_ARG = 143


def extended_format_BUILD_MAP_older(
    opc, instructions: list
) -> Tuple[str, Optional[int]]:
    arg_count = instructions[0].argval
    if arg_count == 0:
        # Note: caller generally handles this when the below isn't right.
        return "{}", instructions[0].offset
    arglist, _, i = get_arglist(instructions, 0, arg_count)
    # from trepan.api import debug; debug()
    if arglist is not None:
        assert isinstance(i, int)
        arg_pairs = [f"{arglist[i]}:{arglist[i+1]}" for i in range(len(arglist), 2)]
        args_str = ", ".join(arg_pairs)
        return "{" + args_str + "}", instructions[i].start_offset
    return "", None


def extended_format_PRINT_ITEM(opc, instructions: list) -> Tuple[str, Optional[int]]:
    instr1 = instructions[1]
    print_arg = get_instruction_arg(instr1)
    return (
        f"print {print_arg},",
        instr1.start_offset,
    )


def extended_format_SLICE_0(opc, instructions: list) -> Tuple[str, Optional[int]]:
    arglist, arg_count, i = get_arglist(instructions, 0, 1)
    if arg_count == 1 and arglist is not None:
        return f"{arglist[0]}[:]", instructions[0].start_offset
    return "", None


def extended_format_SLICE_1(opc, instructions: list) -> Tuple[str, Optional[int]]:
    arglist, arg_count, i = get_arglist(instructions, 0, 2)
    if arg_count == 2 and arglist is not None:
        return f"{arglist[1]}[{arglist[0]}:]", instructions[i].start_offset
    return "", None


def extended_format_SLICE_2(opc, instructions: list) -> Tuple[str, Optional[int]]:
    arglist, arg_count, i = get_arglist(instructions, 0, 2)
    if arg_count == 2 and i is not None and arglist is not None:
        return f"{arglist[1]}[:{arglist[0]}]", instructions[i].start_offset
    return "", None


def extended_format_SLICE_3(opc, instructions: list) -> Tuple[str, Optional[int]]:
    arglist, arg_count, i = get_arglist(instructions, 0, 3)
    if arg_count == 3 and i is not None and arglist is not None:
        arglist = ["" if arg == "None" else arg for arg in arglist]
        return f"{arglist[2]}[{arglist[1]}:{arglist[0]}]", instructions[i].start_offset

    if instructions[0].argval == 0:
        # Degenerate case
        return "set()", instructions[0].start_offset
    return "", None


def format_PRINT_NEWLINE(arg) -> str:
    return 'print "\\n"'


update_arg_fmt_base2x = {
    **opcode_arg_fmt_base,
    **{
        "MAKE_FUNCTION": format_MAKE_FUNCTION_10_27,
        "PRINT_NEWLINE": format_PRINT_NEWLINE,
    },
}

opcode_extended_fmt_base2x = {
    **opcode_extended_fmt_base,
    **{
        # "BUILD_MAP": extended_format_BUILD_MAP_older,
        "PRINT_ITEM": extended_format_PRINT_ITEM,
        "SLICE+0": extended_format_SLICE_0,
        "SLICE+1": extended_format_SLICE_1,
        "SLICE+2": extended_format_SLICE_2,
        "SLICE+3": extended_format_SLICE_3,
    },
}



# (C) Copyright 2019-2021, 2023-2024 by Rocky Bernstein
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
CPython 3.9 bytecode opcodes

This is like Python 3.9's opcode.py with some classification
of stack usage and information for formatting instructions.
"""

from typing import Optional, Tuple

import xdis.opcodes.opcode_38 as opcode_38
from xdis.opcodes.base import (
    binary_op,
    def_op,
    finalize_opcodes,
    init_opdata,
    jabs_op,
    rm_op,
    update_pj3,
)
from xdis.opcodes.format.extended import extended_format_binary_op
from xdis.opcodes.opcode_38 import opcode_arg_fmt38, opcode_extended_fmt38

version_tuple = (3, 9)
python_implementation = "CPython"

loc = locals()

init_opdata(loc, opcode_38, version_tuple)

# fmt: off
# These are removed since 3.8...
rm_op(loc, "BEGIN_FINALLY",       53)
rm_op(loc, "WITH_CLEANUP_START",  81)
rm_op(loc, "WITH_CLEANUP_FINISH", 82)
rm_op(loc, "END_FINALLY",         88)
rm_op(loc, "BUILD_LIST_UNPACK",  149)
rm_op(loc, "BUILD_MAP_UNPACK",   150)
rm_op(loc, "BUILD_MAP_UNPACK_WITH_CALL", 151)
rm_op(loc, "BUILD_TUPLE_UNPACK", 152)
rm_op(loc, "BUILD_SET_UNPACK",   153)
rm_op(loc, "BUILD_TUPLE_UNPACK_WITH_CALL", 158)
rm_op(loc, "CALL_FINALLY",       162)
rm_op(loc, "POP_FINALLY",        163)


# These are new since Python 3.9

#          OP NAME               OPCODE  POP PUSH
#------------------------------------------------
def_op(loc, "RERAISE",                 48,   3, 0)
def_op(loc, "WITH_EXCEPT_START",       49,   0, 1)
def_op(loc, "LOAD_ASSERTION_ERROR",    74,   0, 1)
def_op(loc, "LIST_TO_TUPLE",           82,   1, 1)

binary_op(loc, "IS_OP",               117)
jabs_op(loc, "JUMP_IF_NOT_EXC_MATCH", 121,   2, 0)
binary_op(loc, "CONTAINS_OP",         118,   2, 1)
def_op(loc, "LIST_EXTEND",            162,   2, 1)
def_op(loc, "SET_UPDATE",             163,   2, 1)
def_op(loc, "DICT_MERGE",             164,   2, 1)
def_op(loc, "DICT_UPDATE",            165,   2, 1)

# fmt: on


def extended_format_CONTAINS_OP(
    opc, instructions
) -> Tuple[Optional[str], Optional[int]]:
    instr = instructions[0]
    return extended_format_binary_op(
        opc, instructions, f"%s {format_CONTAINS_OP(instr.arg)} %s"
    )


def extended_format_IS_OP(opc, instructions) -> Tuple[Optional[str], Optional[int]]:
    instr = instructions[0]
    return extended_format_binary_op(
        opc, instructions, f"%s {format_IS_OP(instr.arg)} %s"
    )


def format_CONTAINS_OP(arg):
    return "in" if arg == 0 else "not in"


def format_IS_OP(arg):
    return "is" if arg == 0 else "is not"


opcode_arg_fmt = opcode_arg_fmt39 = {
    **opcode_arg_fmt38,
    **{
        "CONTAINS_OP": format_CONTAINS_OP,
        "IS_OP": format_IS_OP,
    },
}

opcode_extended_fmt = opcode_extended_fmt39 = {
    **opcode_extended_fmt38.copy(),
    **{
        "CONTAINS_OP": extended_format_CONTAINS_OP,
        "IS_OP": extended_format_IS_OP,
    },
}

update_pj3(globals(), loc)
finalize_opcodes(loc)
